package node

//go:generate go-bindata --nocompress --pkg node static/... templates/...

import (
	"bytes"
	"fmt"
	"io"
	"log"
	"net/http"
	"net/url"
	"strconv"
	"strings"

	"git.autistici.org/ale/autoradio"
	pb "git.autistici.org/ale/autoradio/proto"
	"github.com/alecthomas/template"
	assetfs "github.com/elazarl/go-bindata-assetfs"
)

func NewHTTP(n *Node, lb *loadBalancer, icecastPort int, domain string) http.Handler {
	mux := http.NewServeMux()
	tpl := mustParseEmbeddedTemplates()

	// Serve /static/ from builtin assets.
	mux.Handle("/static/", http.StripPrefix("/static/", http.FileServer(&assetfs.AssetFS{
		Asset:     Asset,
		AssetDir:  AssetDir,
		AssetInfo: AssetInfo,
		Prefix:    "static",
	})))

	// Requests for /_stream/ go straight to the local Icecast.
	proxyHandler := http.StripPrefix(autoradio.IcecastMountPrefix,
		withMount(n, func(m *pb.Mount, w http.ResponseWriter, r *http.Request) {
			doIcecastProxy(w, r, &url.URL{
				Scheme: "http",
				Host:   fmt.Sprintf("localhost:%d", icecastPort),
				Path:   autoradio.MountPathToIcecastPath(m.Path),
			}, m.Path)
		}))

	// redirectHandler serves different kinds of redirects, either
	// an M3U or a HTTP redirect, for the public stream URLs (ones
	// without the /_stream prefix).
	redirectHandler := withMount(n, func(m *pb.Mount, w http.ResponseWriter, r *http.Request) {
		serveRedirect(n.lb, m, w, r)
	})

	// sourceHandler deals with SOURCE requests to the public
	// stream URL, which are forwarded straight to the master
	// Icecast node.
	sourceHandler := withMount(n, func(m *pb.Mount, w http.ResponseWriter, r *http.Request) {
		serveSource(n, m, w, r)
	})

	streamPrefixSlash := autoradio.IcecastMountPrefix + "/"
	mux.HandleFunc("/", func(w http.ResponseWriter, r *http.Request) {
		switch {
		case r.Method == "GET" && strings.HasPrefix(r.URL.Path, streamPrefixSlash):
			proxyHandler.ServeHTTP(w, r)
		case r.Method == "SOURCE" || r.Method == "PUT":
			sourceHandler.ServeHTTP(w, r)
		case r.Method == "GET" && (r.URL.Path == "" || r.URL.Path == "/"):
			serveStatusPage(n, w, r, tpl, domain)
		case r.Method == "GET":
			redirectHandler.ServeHTTP(w, r)
		default:
			http.NotFound(w, r)
		}
	})

	return mux
}

// Request wrapper that passes a Mount along with the HTTP
// request. The mount path is just the URL path without an eventual
// .m3u extension.
func withMount(n *Node, f func(*pb.Mount, http.ResponseWriter, *http.Request)) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		mountPath := strings.TrimSuffix(r.URL.Path, ".m3u")
		mount, ok := n.mounts.GetMount(mountPath)
		if !ok {
			http.NotFound(w, r)
			return
		}
		f(mount, w, r)
	})
}

func serveSource(n *Node, mount *pb.Mount, w http.ResponseWriter, r *http.Request) {
	// Can't connect sources to relay streams.
	if mount.IsRelay() {
		log.Printf("source connection to relay stream %s", mount.Path)
		http.Error(w, "Stream is relayed, no source connections allowed", http.StatusBadRequest)
		return
	}

	targetAddr := n.leaderAddr()
	if targetAddr == "" {
		http.Error(w, "No leader", http.StatusServiceUnavailable)
		return
	}

	doIcecastProxy(w, r, &url.URL{
		Scheme: "http",
		Host:   targetAddr,
		Path:   autoradio.MountPathToIcecastPath(mount.Path),
	}, mount.Path)
}

func serveRedirect(lb *loadBalancer, mount *pb.Mount, w http.ResponseWriter, r *http.Request) {
	// Redirect the user to a final target, depending on the load
	// balancing algorithm's decision.  This enforces the
	// 1:1 mapping between Icecasts and frontends.
	targetNode := lb.chooseNode(&httpRequestContext{r})
	if targetNode == nil {
		http.Error(w, "No nodes available", http.StatusServiceUnavailable)
		return
	}

	targetAddr := randomIPWithMatchingProtocol(targetNode.parsedAddrs, r.RemoteAddr)
	if targetAddr == "" {
		// This should not happen if the protocol filter in
		// the load balancer evaluation did its job properly.
		http.Error(w, "No nodes available with the right IP protocol", http.StatusServiceUnavailable)
		return
	}

	if strings.HasSuffix(r.URL.Path, ".m3u") {
		sendM3U(w, r)
	} else {
		targetURL := fmt.Sprintf("http://%s%s", targetAddr, autoradio.MountPathToIcecastPath(mount.Path))
		sendRedirect(w, r, targetURL)
	}
}

// Serve a M3U response. This simply points back at the stream
// redirect handler by dropping the .m3u suffix in the request URL.
func sendM3U(w http.ResponseWriter, r *http.Request) {
	m3u := strings.TrimSuffix(r.URL.String(), ".m3u") + "\n"
	w.Header().Set("Content-Length", strconv.Itoa(len(m3u)))
	w.Header().Set("Content-Type", "audio/x-mpegurl")
	addDefaultHeaders(w)
	io.WriteString(w, m3u) //nolint
}

func sendRedirect(w http.ResponseWriter, r *http.Request, targetURL string) {
	// Firefox apparently caches redirects regardless of
	// the status code, so we have to add some quite
	// aggressive cache-busting headers. We serve a status
	// code of 307 to HTTP/1.1 clients, 302 otherwise.
	w.Header().Set("Cache-Control", "max-age=0,no-cache,no-store")
	w.Header().Set("Pragma", "no-cache")
	w.Header().Set("Expires", "-1")
	code := http.StatusFound
	if r.ProtoMinor == 1 {
		code = http.StatusTemporaryRedirect
	}
	http.Redirect(w, r, targetURL, code)
}

func serveStatusPage(n *Node, w http.ResponseWriter, r *http.Request, tpl *template.Template, domain string) {
	nodes := n.lb.getNodes()
	ms := mountsToStatus(n.mounts.GetMounts(), nodes)
	ctx := struct {
		Domain string
		Nodes  []*nodeInfo
		Mounts []*mountStatus
	}{domain, nodes, ms}

	var buf bytes.Buffer
	if err := tpl.ExecuteTemplate(&buf, "index.html", ctx); err != nil {
		log.Printf("error rendering status page: %v", err)
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	w.Header().Set("Content-Type", "text/html; charset=utf-8")
	w.Header().Set("Content-Length", strconv.Itoa(buf.Len()))
	addDefaultHeaders(w)
	w.Write(buf.Bytes()) //nolint
}

func addDefaultHeaders(w http.ResponseWriter) {
	w.Header().Set("Expires", "-1")
	w.Header().Set("Cache-Control", "no-cache,no-store")
}

// Parse the templates that are embedded with the binary (in bindata.go).
func mustParseEmbeddedTemplates() *template.Template {
	root := template.New("")
	files, err := AssetDir("templates")
	if err != nil {
		panic(err)
	}
	for _, f := range files {
		b, err := Asset("templates/" + f)
		if err != nil {
			log.Fatalf("could not read embedded template %s: %v", f, err)
		}
		if _, err := root.New(f).Parse(string(b)); err != nil {
			log.Fatalf("error parsing template %s: %v", f, err)
		}
	}
	return root
}
