package node

import (
	"context"
	"log"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/util"
	"git.autistici.org/ale/autoradio/client"
	"git.autistici.org/ale/autoradio/coordination/election"
	"git.autistici.org/ale/autoradio/coordination/presence"
	pb "git.autistici.org/ale/autoradio/proto"
	"go.etcd.io/etcd/clientv3"
	"go.etcd.io/etcd/clientv3/concurrency"
)

var (
	statusUpdateInterval = 1 * time.Second

	sessionTTL = 5 // seconds
)

// Icecast is the interface used to manage the local Icecast daemon.
type Icecast interface {
	GetStatus() ([]*pb.IcecastMount, bool)
	Update(context.Context, []*pb.Mount, bool, string) error
}

// A Node controls an Icecast instance and participates in various
// presence/consensus protocols in order to be part of the autoradio
// cluster. Its only communications are with Icecast and the
// frontends. A Node comes with its own etcd Session, that it
// maintains until the controlling Context is canceled.
type Node struct {
	name string

	mounts      client.MountConfig
	watcher     *election.ElectionWatcher
	updateCh    chan struct{}
	ice         Icecast
	publicPeers *presence.EndpointSet
	lb          *loadBalancer

	ctx    context.Context
	cancel context.CancelFunc
}

func (n *Node) updateIcecastThread() {
	for range n.updateCh {
		elState := n.watcher.State()
		if elState.State == election.StateUnknown {
			log.Printf("not reloading Icecast because leadership status is unknown")
			continue
		}

		err := n.ice.Update(n.ctx, n.mounts.GetMounts(),
			(elState.State == election.StateLeader),
			elState.Leader.Addrs[0])
		if err != nil {
			log.Printf("error reloading Icecast: %v", err)
		}
		time.Sleep(3 * time.Second)
	}
}

func (n *Node) updateIcecast() {
	select {
	case n.updateCh <- struct{}{}:
	default:
	}
}

func (n *Node) getStatus() *pb.Status {
	iceMounts, iceOk := n.ice.GetStatus()
	ns := pb.Status{
		Name:          n.name,
		Timestamp:     uint64(time.Now().UTC().UnixNano()),
		IcecastOk:     iceOk,
		IcecastMounts: iceMounts,
		BwUsage:       0,
		MaxListeners:  1000,
	}
	return &ns
}

func (n *Node) leaderAddr() string {
	leader := n.watcher.State().Leader
	if leader == nil || len(leader.Addrs) == 0 {
		return ""
	}
	return leader.Addrs[0]
}

// New returns a new Node.
func New(cli *clientv3.Client, nodeID, publicAddrs, peerAddr string, ice Icecast, lbSpec string) (*Node, error) {
	// The top-level etcd Session controls the lifetime of the
	// ephemeral etcd keys that store presence- and status-related
	// data for this Node.
	session, err := concurrency.NewSession(cli, concurrency.WithTTL(sessionTTL))
	if err != nil {
		return nil, err
	}

	// Create a top-level Context that can be canceled when Stop
	// is called. This Context controls the lifetime of the Node
	// itself, and all the associated background goroutines:
	// canceling it immediately terminates all outbound requests.
	ctx, cancel := context.WithCancel(context.Background())
	go func() {
		// Oops, the session is gone, stop everything.
		<-session.Done()
		cancel()
	}()

	n := &Node{
		ice:      ice,
		ctx:      ctx,
		cancel:   cancel,
		name:     nodeID,
		updateCh: make(chan struct{}, 1),
	}

	// The runtime configuration is just a list of
	// mounts. Synchronize it with etcd, wait until it is ready,
	// and restart Icecast whenever it changes.
	configReady, notify := client.WatchConfig(ctx, cli, &n.mounts)
	go func() {
		for range notify {
			n.updateIcecast()
		}
	}()
	<-configReady

	// Register the Icecast endpoints.
	_, err = presence.Register(
		ctx,
		session,
		nodeID,
		presence.NewRegistration(autoradio.StatusEndpointPrefix, peerAddr, autoradio.IcecastPort),
		presence.NewRegistration(autoradio.PublicEndpointPrefix, publicAddrs, autoradio.IcecastPort),
	)
	if err != nil {
		cancel()
		return nil, err
	}

	// Watch the authoritative list of peer nodes.
	n.publicPeers = presence.WatchEndpoints(ctx, cli, autoradio.PublicEndpointPrefix)

	// Run a leader election protocol advertising the peer address
	// of our Icecast daemon.
	icecastEP, err := pb.ParseEndpoint(nodeID, peerAddr, autoradio.IcecastPort)
	if err != nil {
		cancel()
		return nil, err
	}
	// We participate in the leader election, but then do nothing
	// special once we're the leader, just wait for
	// cancellation. All the state transitions are managed by the
	// election Watcher, which triggers an Icecast update.
	el := election.New(session, autoradio.IcecastElectionPrefix, icecastEP)
	go func() {
		err := el.Run(ctx, election.WaitForever)

		// If the election runner's outer retry loop
		// terminates, we are unable to talk to etcd to
		// re-acquire leadership. In this case we should abort
		// the Node, to prevent ending up in a state where all
		// nodes are running but the cluster has no leader.
		if err != nil && err != context.Canceled {
			log.Printf("leader election aborted: %v", err)
		}
		cancel()
	}()

	n.watcher = election.NewWatcher(ctx, el)
	go func() {
		for range n.watcher.Notify() {
			log.Printf("leader election state: %+v", n.watcher.State())
			n.updateIcecast()
		}
	}()

	// Start the background thread that updates Icecast whenever
	// something changes.
	go n.updateIcecastThread()

	// Start the status reporter that periodically sends our
	// status to the frontends.
	statusMgr := newStatusManager(ctx, cli)
	go util.RunCron(ctx, statusUpdateInterval, func(_ context.Context) {
		statusMgr.update(n.getStatus())
	})

	// Create the loadBalancer that runs within the node.
	n.lb, err = newLoadBalancer(ctx, n.publicPeers, statusMgr, lbSpec)
	if err != nil {
		cancel()
		return nil, err
	}

	return n, nil
}

func (n *Node) Stop() {
	n.cancel()
}

func (n *Node) Wait() {
	<-n.ctx.Done()
}
